#include "Project.h"

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef USE_FM_SYNTH

FM_form form;
FM_synth fm;




// ------------------------------------------------------------------------------
// FRAMEWORK DOCUMENTATION
//
// NOTE: Resets wave form to default values. *Beep*
//
// ------------------------------------------------------------------------------
//
void FM_reset_form()
{
 form.wave_type     = 0;
 form.filter_on     = 0;
 form.base_freq     = 0.3f;
 form.freq_limit    = 0.0f;
 form.freq_ramp     = 0.0f;
 form.freq_dramp    = 0.0f;
 form.duty          = 0.0f;
 form.duty_ramp     = 0.0f;
 form.vib_strength  = 0.0f;
 form.vib_speed     = 0.0f;
 form.vib_delay     = 0.0f;
 form.env_attack    = 0.0f;
 form.env_sustain   = 0.3f;
 form.env_decay     = 0.4f;
 form.env_punch     = 0.0f;
 form.lpf_resonance = 0.0f;
 form.lpf_freq      = 1.0f;
 form.lpf_ramp      = 0.0f;
 form.hpf_freq      = 0.0f;
 form.hpf_ramp      = 0.0f;
 form.pha_offset    = 0.0f;
 form.pha_ramp      = 0.0f;
 form.repeat_speed  = 0.0f;
 form.arp_speed     = 0.0f;
 form.arp_mod       = 0.0f;

 fm.master_vol      = 0.05f;
 fm.sound_vol       = 0.5f;
 fm.wav_bits        = 16;
 fm.wav_freq        = 44100;
 fm.filesample      = 0.0f;
 fm.fileacc         = 0;
 fm.playing_sample  = 0;

 return;
}





// ------------------------------------------------------------------------------
// FRAMEWORK DOCUMENTATION
//
// NOTE: Loads a SFXR-compatible file from disk
//
// NOTE: Returns [1] to indicate success, and [0] to indicate failure
//
// ------------------------------------------------------------------------------
//
bool FM_load_form(const char *filename)
{
 int version = 0;

 FILE *file = fopen(filename, "rb");
 if(!file) {
     log("[ERROR] - %s was not accessible when loading FM form", filename);
     return(0);
           }

 // Version check. Bail if it's not something we recognize
 //
 fread(&version, 1, sizeof(int), file);
 if(version != 100 && version != 101 && version != 102) {
     log("[ERROR] - Version reported by file (%d) isn't 100, 101, or 102", version);
     fclose(file);
     return(0);
                                                        }

 fread(&form.wave_type,     1, sizeof(int),   file);

 // This param only shows up in v102 files
 //
 fm.sound_vol = 0.5f;
 if(version == 102) { fread(&fm.sound_vol, 1, sizeof(float), file); }

 fread(&form.base_freq,     1, sizeof(float), file);
 fread(&form.freq_limit,    1, sizeof(float), file);
 fread(&form.freq_ramp,     1, sizeof(float), file);

 // This param only shows up in v101 files
 //
 if(version >= 101) { fread(&form.freq_dramp, 1, sizeof(float), file); }

 fread(&form.duty,          1, sizeof(float), file);
 fread(&form.duty_ramp,     1, sizeof(float), file);

 fread(&form.vib_strength,  1, sizeof(float), file);
 fread(&form.vib_speed,     1, sizeof(float), file);
 fread(&form.vib_delay,     1, sizeof(float), file);

 fread(&form.env_attack,    1, sizeof(float), file);
 fread(&form.env_sustain,   1, sizeof(float), file);
 fread(&form.env_decay,     1, sizeof(float), file);
 fread(&form.env_punch,     1, sizeof(float), file);

 fread(&form.filter_on,     1, sizeof(bool),  file);
 fread(&form.lpf_resonance, 1, sizeof(float), file);
 fread(&form.lpf_freq,      1, sizeof(float), file);
 fread(&form.lpf_ramp,      1, sizeof(float), file);
 fread(&form.hpf_freq,      1, sizeof(float), file);
 fread(&form.hpf_ramp,      1, sizeof(float), file);

 fread(&form.pha_offset,    1, sizeof(float), file);
 fread(&form.pha_ramp,      1, sizeof(float), file);
 fread(&form.repeat_speed,  1, sizeof(float), file);

 if(version >= 101) {
     fread(&form.arp_speed, 1, sizeof(float), file);
     fread(&form.arp_mod,   1, sizeof(float), file);
                    }

 fclose(file);

 return(1);
}





// ------------------------------------------------------------------------------
// FRAMEWORK DOCUMENTATION
//
// NOTE: Saves current form to a SFXR-compatible file
//
// NOTE: Returns [1] to indicate success, and [0] to indicate failure
//
// ------------------------------------------------------------------------------
//
bool FM_save_form(const char *filename)
{
 int version = 102; // May as well keep this

 FILE *file = fopen(filename, "wb");
 if(!file) {
     log("[ERROR] - Couldn't save %s", filename);
     return(0);
           }

 fwrite(&version,            1, sizeof(int),   file);
 fwrite(&form.wave_type,     1, sizeof(int),   file);
 fwrite(&fm.sound_vol,       1, sizeof(float), file);

 fwrite(&form.base_freq,     1, sizeof(float), file);
 fwrite(&form.freq_limit,    1, sizeof(float), file);
 fwrite(&form.freq_ramp,     1, sizeof(float), file);
 fwrite(&form.freq_dramp,    1, sizeof(float), file);
 fwrite(&form.duty,          1, sizeof(float), file);
 fwrite(&form.duty_ramp,     1, sizeof(float), file);

 fwrite(&form.vib_strength,  1, sizeof(float), file);
 fwrite(&form.vib_speed,     1, sizeof(float), file);
 fwrite(&form.vib_delay,     1, sizeof(float), file);

 fwrite(&form.env_attack,    1, sizeof(float), file);
 fwrite(&form.env_sustain,   1, sizeof(float), file);
 fwrite(&form.env_decay,     1, sizeof(float), file);
 fwrite(&form.env_punch,     1, sizeof(float), file);

 fwrite(&form.filter_on,     1, sizeof(bool),  file);
 fwrite(&form.lpf_resonance, 1, sizeof(float), file);
 fwrite(&form.lpf_freq,      1, sizeof(float), file);
 fwrite(&form.lpf_ramp,      1, sizeof(float), file);
 fwrite(&form.hpf_freq,      1, sizeof(float), file);
 fwrite(&form.hpf_ramp,      1, sizeof(float), file);

 fwrite(&form.pha_offset,    1, sizeof(float), file);
 fwrite(&form.pha_ramp,      1, sizeof(float), file);

 fwrite(&form.repeat_speed,  1, sizeof(float), file);

 fwrite(&form.arp_speed,     1, sizeof(float), file);
 fwrite(&form.arp_mod,       1, sizeof(float), file);

 fclose(file);

 return(1);
}





// ------------------------------------------------------------------------------
// FRAMEWORK DOCUMENTATION
//
// NOTE: This resets the sample (but NOT the form on which the sample is based)
//
// ------------------------------------------------------------------------------
//
void FM_reset_sample(bool restart)
{
 int i = 0;

 if(!restart) { fm.phase = 0; }

 // I did not aggressively reformat this because it'd be easy
 // to screw up the order of operation on many of these lines
 //
 fm.fperiod      = 100.0 / (form.base_freq * form.base_freq + 0.001);
 fm.period       = (int)fm.fperiod;
 fm.fmaxperiod   = 100.0 / (form.freq_limit * form.freq_limit + 0.001);
 fm.fslide       = 1.0 - pow((double)form.freq_ramp, 3.0) * 0.01;
 fm.fdslide      = -pow((double)form.freq_dramp, 3.0) * 0.000001;
 fm.square_duty  = 0.5f - form.duty * 0.5f;
 fm.square_slide = -form.duty_ramp * 0.00005f;

 if(form.arp_mod >= 0.0f) { fm.arp_mod = 1.0 - pow((double)form.arp_mod, 2.0) * 0.9;  }
 else                     { fm.arp_mod = 1.0 + pow((double)form.arp_mod, 2.0) * 10.0; }

 fm.arp_time  = 0;
 fm.arp_limit = (int)(pow(1.0f - form.arp_speed, 2.0f) * 20000 + 32);

 if(form.arp_speed == 1.0f) { fm.arp_limit = 0; }

 if(!restart) {
     // reset filter
     //
     fm.fltp   = 0.0f;
     fm.fltdp  = 0.0f;
     fm.fltw   = pow(form.lpf_freq, 3.0f) * 0.1f;
     fm.fltw_d = 1.0f + form.lpf_ramp * 0.0001f;
     fm.fltdmp = 5.0f / (1.0f + pow(form.lpf_resonance, 2.0f) * 20.0f) * (0.01f + fm.fltw);

     if(fm.fltdmp > 0.8f) { fm.fltdmp = 0.8f; }

     fm.fltphp  = 0.0f;
     fm.flthp   = pow(form.hpf_freq, 2.0f) * 0.1f;
     fm.flthp_d = 1.0 + form.hpf_ramp * 0.0003f;

     // reset vibrato
     //
     fm.vib_phase = 0.0f;
     fm.vib_speed = pow(form.vib_speed, 2.0f) * 0.01f;
     fm.vib_amp   = (form.vib_strength * 0.5f);

     // reset envelope
     //
     fm.env_vol       = 0.0f;
     fm.env_stage     = 0;
     fm.env_time      = 0;
     fm.env_length[0] = (int)(form.env_attack  * form.env_attack  * 100000.0f);
     fm.env_length[1] = (int)(form.env_sustain * form.env_sustain * 100000.0f);
     fm.env_length[2] = (int)(form.env_decay   * form.env_decay   * 100000.0f);

     fm.fphase = pow(form.pha_offset, 2.0f) * 1020.0f;
     if(form.pha_offset<0.0f) { fm.fphase = -fm.fphase; }

     fm.fdphase = pow(form.pha_ramp, 2.0f) * 1.0f;
     if(form.pha_ramp < 0.0f) { fm.fdphase = -fm.fdphase; }

     fm.iphase = abs((int)fm.fphase);
     fm.ipp = 0;
     for(i=0; i<1024; i++) { fm.phaser_buffer[i] = 0.0f;                                        }
     for(i=0; i<32; i++)   { fm.noise_buffer[i] = (float)(quick_rand(20000) - 10000) / 10000.0; }

     fm.rep_time = 0;
     fm.rep_limit = (int)(pow(1.0f - form.repeat_speed, 2.0f) * 20000 + 32);
     if(form.repeat_speed == 0.0f) { fm.rep_limit = 0; }
	          }

 return;
}






// ------------------------------------------------------------------------------
// FRAMEWORK DOCUMENTATION
//
// NOTE: Rejiggers some internal SFXR variables. Should not be called by
//       external code -- there's no reason to.
//
// ------------------------------------------------------------------------------
//
void FM_play_sample()
{

 FM_reset_sample(0);
 fm.playing_sample = 1;

 return;
}






// ------------------------------------------------------------------------------
// FRAMEWORK DOCUMENTATION
//
// NOTE: This synthesizes a sample from the current wave form. If NULL is
//       passed for [BUFFER] then no actual rendering will occur.
//
// ------------------------------------------------------------------------------
//
void FM_synth(int length, unsigned char* buffer, int offset)
{
 int i  = 0;
 int j  = 0;
 int si = 0;


 for(i=0; i<length; i++) {
     if(!fm.playing_sample) { break; }

     fm.rep_time++;
     if(fm.rep_limit!=0 && fm.rep_time>=fm.rep_limit) {
         fm.rep_time=0;
         FM_reset_sample(1);
                                                      }

     // frequency envelopes/arpeggios
     //
     fm.arp_time++;
     if(fm.arp_limit != 0 && fm.arp_time >= fm.arp_limit) {
         fm.arp_limit = 0;
         fm.fperiod *= fm.arp_mod;
                                                          }

     fm.fslide += fm.fdslide;
     fm.fperiod *= fm.fslide;

     if(fm.fperiod > fm.fmaxperiod) {
         fm.fperiod = fm.fmaxperiod;
         if(form.freq_limit > 0.0f) { fm.playing_sample = 0; }
                                    }

     float rfperiod = fm.fperiod; // don't like this. Revise
     if(fm.vib_amp > 0.0f) {
         fm.vib_phase += fm.vib_speed;
         rfperiod = fm.fperiod * (1.0 + sin(fm.vib_phase) * fm.vib_amp);
                           }

     fm.period = (int)rfperiod;
     if(fm.period < 8) { fm.period = 8; }

     fm.square_duty += fm.square_slide;
     if(fm.square_duty < 0.0f) { fm.square_duty = 0.0f; }
     if(fm.square_duty > 0.5f) { fm.square_duty = 0.5f; }

     // volume envelope
     //
     fm.env_time++;
     if(fm.env_time > fm.env_length[fm.env_stage]) {
         fm.env_time = 0;
         fm.env_stage++;
         if(fm.env_stage == 3) { fm.playing_sample = 0; }
                                                   }

     if(fm.env_stage==0) { fm.env_vol = (float)fm.env_time / fm.env_length[0]; }
     if(fm.env_stage==1) { fm.env_vol = 1.0f + pow(1.0f - (float)fm.env_time / fm.env_length[1], 1.0f) * 2.0f * form.env_punch; }
     if(fm.env_stage==2) { fm.env_vol = 1.0f - (float)fm.env_time / fm.env_length[2]; }

     // phaser step
     //
     fm.fphase += fm.fdphase;
     fm.iphase = abs((int)fm.fphase);
     if(fm.iphase > 1023) { fm.iphase = 1023; }

     if(fm.flthp_d != 0.0f) {
         fm.flthp *= fm.flthp_d;
         if(fm.flthp < 0.00001f) { fm.flthp = 0.00001f; }
         if(fm.flthp > 0.1f)     { fm.flthp = 0.1f;     }
                            }

     float ssample = 0.0f; // Also don't like this. Revise

     for(si=0; si<8; si++) { // 8x supersampling
         float sample = 0.0f; // Goddamnit. More of this bs.
         fm.phase++;
         if(fm.phase >= fm.period) {
             fm.phase %= fm.period;
             if(form.wave_type == 3) {
                 for(j=0; j<32; j++) { fm.noise_buffer[j] = (float)(quick_rand(20000) - 10000) / 10000.0; }
                                     }
                                   }

         // base waveform
         //
         float fp = (float)fm.phase/fm.period; // I give up.

         switch(form.wave_type) {
             case 0: // square
                     if(fp<fm.square_duty) { sample = 0.5f;  }
                     else                  { sample = -0.5f; }
                     break;

             case 1: // sawtooth
                     sample = 1.0f - fp * 2;
                     break;

             case 2: // sine
                     sample = (float)sin(fp * 2 * fm_PI);
                     break;

             case 3: // noise
                     sample = fm.noise_buffer[fm.phase * 32 / fm.period];
                     break;
                                }

         // lp filter
         //
         float pp = fm.fltp;
         fm.fltw *= fm.fltw_d;
         if(fm.fltw < 0.0f) { fm.fltw = 0.0f; }
         if(fm.fltw > 0.1f) { fm.fltw = 0.1f; }

         if(form.lpf_freq != 1.0f) {
             fm.fltdp += (sample - fm.fltp) * fm.fltw;
             fm.fltdp -= fm.fltdp * fm.fltdmp;
                                   }

         else {
             fm.fltp  = sample;
             fm.fltdp = 0.0f;
              }

         fm.fltp += fm.fltdp;

         // hp filter
         //
         fm.fltphp += fm.fltp - pp;
         fm.fltphp -= fm.fltphp * fm.flthp;
         sample =  fm.fltphp;

         // phaser
         //
         fm.phaser_buffer[fm.ipp & 1023] = sample;
         sample += fm.phaser_buffer[(fm.ipp - fm.iphase + 1024) & 1023];
         fm.ipp = (fm.ipp + 1) & 1023;

         // final accumulation and envelope application
         //
         ssample += sample * fm.env_vol;
		                   }

     ssample = ssample / 8 * fm.master_vol;
     ssample *= 2.0f * fm.sound_vol;

     if(buffer != NULL) {
         // quantize depending on format
         // accumulate/count to accomodate variable sample rate?
         //
         ssample *= 4.0f; // arbitrary gain to get reasonable output volume...
         if(ssample > 1.0f)  { ssample = 1.0f;  }
         if(ssample < -1.0f) { ssample = -1.0f; }

         fm.filesample += ssample;
         fm.fileacc++;
         if(fm.wav_freq == 44100 || fm.fileacc == 2) {
             fm.filesample /= fm.fileacc;
             fm.fileacc = 0;

             short isample = (short)(fm.filesample * 32000);

             #ifdef FM_MONO_INPUT
             ((unsigned char *)buffer)[offset + (i * 2) + 0] = (isample & 0xFF);
             ((unsigned char *)buffer)[offset + (i * 2) + 1] = (isample >> 8) & 0xFF;
             #endif

             #ifndef FM_MONO_INPUT
             ((unsigned char *)buffer)[offset + (i * 4) + 0] = (isample & 0xFF);
             ((unsigned char *)buffer)[offset + (i * 4) + 1] = (isample >> 8) & 0xFF;
             ((unsigned char *)buffer)[offset + (i * 4) + 2] = (isample & 0xFF);
             ((unsigned char *)buffer)[offset + (i * 4) + 3] = (isample >> 8) & 0xFF;
             #endif

             fm.filesample = 0.0f;
                                                     }

                        }

     fm.file_sampleswritten++;
	                     }

 return;
}





// NOTE: This searches the resource directory and automatically loads
//       any SFS files it finds matching a XXX.SFS pattern. It then
//       processes them and loads them into the sample cache matching
//       the filename. An error will be logged if that slot is
//       occupied or unavailable due to MAX_SAMPLES being too low
//
void FM_load_assets()
{
 int  i = 0;
 char name[256];

 // We could scan the directory and build a file list, but
 // this might end up being more platform-agnostic.
 //
 for(i=0; i<MAX_SFX; i++) {
     memset(&name[0], 0, 256);
     sprintf(&name[0], "sfx/%03d.sfs", i);

     if(!access(&name[0], F_OK)) {
         FM_reset_form();
         FM_load_form(&name[0]);
         FM_reset_sample(0);

         sfx[i] = FM_export_to_Sound();
                                 }
                          }

 return;
}





// NOTE: This takes a rendered waveform and stores it in a buffer
//       in WAV format, exactly as if it were to be written to a file.
//       We can then either save it to a file, or point Raylib's
//       LoadWaveFromMemory() and LoadSoundFromWave() to turn it into
//       a playable "Sound".
//
Sound FM_export_to_Sound()
{
 const char *header_txt[] = {
 "RIFF",
 "WAVE",
 "fmt ",
 "data"
 };

 unsigned char *buffer;
 int offset = 0;
 int buffer_size = 0;


 // First pass determines size of synth sample (for use in sizing the buffer)
 //
 fm.file_sampleswritten = 0;
 fm.filesample          = 0.0f;
 fm.fileacc             = 0;

 FM_play_sample();

 while(fm.playing_sample) {
     FM_synth(256, NULL, offset); // N.B. the NULL
     #ifdef FM_MONO_INPUT
     offset += 512; // 512 for mono
     #endif

     #ifndef FM_MONO_INPUT
     offset += 1024; // 1024 for stereo
     #endif
                          }


 // Allocate a buffer to hold the stream
 //
 unsigned int dword  = 0;
 unsigned short word = 0;

 buffer_size = 44 + (fm.file_sampleswritten * (sizeof(short) * 2));
 buffer = (unsigned char *)malloc(buffer_size);

 log("buffer size = %d", buffer_size);





 // Construct a WAV header in memory, and append synth to it
 // Simple WAV header is 44 bytes, followed by the raw data
 //
 memcpy(buffer, header_txt[0], 4); // RIFF

 dword = 0; // remaining file size (will fill in below)
 memcpy(buffer + 4, (char*)&dword, sizeof(unsigned int));

 memcpy(buffer + 8,  header_txt[1], 4); // WAVE
 memcpy(buffer + 12, header_txt[2], 4); // fmt

 dword = 16;
 memcpy(buffer + 16, (char*)&dword, sizeof(unsigned int));

 word = 1;
 memcpy(buffer + 20, (char*)&word, sizeof(unsigned short));

 word = 1;
 memcpy(buffer + 22, (char*)&word, sizeof(unsigned short));

 dword = fm.wav_freq;
 memcpy(buffer + 24, (char*)&dword, sizeof(unsigned int));

 dword = fm.wav_freq * fm.wav_bits / 8;
 memcpy(buffer + 28, (char*)&dword, sizeof(unsigned int));

 word = fm.wav_bits / 8;
 memcpy(buffer + 32, (char*)&word, sizeof(unsigned short));

 word = fm.wav_bits;
 memcpy(buffer + 34, (char*)&word, sizeof(unsigned short));

 memcpy(buffer+36, header_txt[3], 4); // data

 dword = 0; // Placeholder for chunk size (will fill in below)
 memcpy(buffer+40, (char*)&dword, sizeof(unsigned int));


 /// ACTUAL RENDERED DATA GOES HERE

 // Second pass is the real deal
 //
 fm.file_sampleswritten = 0;
 fm.filesample          = 0.0f;
 fm.fileacc             = 0;
 offset                 = 0;

 FM_play_sample();

 while(fm.playing_sample) {
     FM_synth(256, (unsigned char *)buffer, offset+44);
     #ifdef FM_MONO_INPUT
     offset += 512; // 512 for mono
     #endif

     #ifndef FM_MONO_INPUT
     offset += 1024; // 1024 for stereo
     #endif
                          }

 /// ACTUAL RENDERED DATA GOES HERE




 // Fill in placeholder data above
 //
 dword = 40 + fm.file_sampleswritten * fm.wav_bits / 8;
 memcpy(buffer + 4, (unsigned char*)&dword, sizeof(unsigned int));

 dword = (fm.file_sampleswritten * fm.wav_bits / 8) + 4;
 memcpy(buffer+40, (char*)&dword, sizeof(unsigned int));

 log("Rendered wave");


 /* UNCOMMENT IF YOU WANT TO DUMP THE WAVE TO DISK
 FILE * fp;
 fp = fopen ("file.wav", "wb+");
 fwrite(buffer, buffer_size, 1, fp);
 fclose(fp);
 */


 Wave wave   = LoadWaveFromMemory(".wav", buffer, buffer_size);
 Sound sound = LoadSoundFromWave(wave);

 // Clean up our mess
 //
 UnloadWave(wave);
 free(buffer);

 return(sound);
}

#endif





