#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "raylib.h"
#include "fastprng.h"

#ifdef USE_ARMCADIA
    #include "ARMCADIA.H"
#endif


#if defined(PLATFORM_WEB)
    #include <emscripten/emscripten.h>
#endif


/// CORE FEATURES TO COMPILE/LINK
///
#define USE_MATH                                   // General purpose math routines
#define USE_UTILITY                                // General purpose routines
#define USE_ANIMATIONS                             // Basic sprite animation routines
#define USE_RINGS                                  // Ring animations
#define USE_TEXTBOX                                // Textbox support
#define USE_HIGH_SCORE                             // Basic High-score support
#define USE_TIPS                                   // Tutorial/Tooltip support (requires textbox)
#define USE_BULLETIN                               // Scrolling bulletin at top/bottom screen
#define USE_LOG                                    // Logfile and console echo support
#define USE_FM_SYNTH                               // SFXR sound Rendering support
//#define USE_ARMCADIA                             // Use Armcadia for dynamic tentacles
//#define USE_RTMIDI                               // MIDI controller support


/// PROJECT-CONFIG
///
#define WINDOW_TITLE    "CATATONIC"
#define SCREEN_W               1280
#define SCREEN_H               1024
#define GLSL_VERSION            330
#define OS_SCALING              1.0
#define BASE_TICKS               60
#define PIX                     4.0

// #define SKIP_INTRO
// #define ENABLE_DEBUG_KEYS
#define NO_ESCAPE_TO_OS
#define HIDE_MOUSE_CURSOR
#define LOG_TO_CONSOLE

#define MAX_ACTORS             1600
#define MAX_ANIMATIONS           16
#define MAX_FRAMES               16
#define MAX_SFX                 256
#define MAX_MUSIC                 8

#define MAIN_MENU_OPTIONS         3
#define MAIN_MENU_BAR_COLOR   270.0
#define MAIN_MENU_BAR_RANGE    20.0
#define MAIN_MENU_BAR_ALPHA    0.35
#define MAIN_MENU_BAR_SPACING    10
#define MAIN_MENU_BAR_SIZE       14

#define SUMMARY_ITEMS             4
#define SUMMARY_ITEM_SIZE       3.0
#define SUMMARY_ITEM_COLOR     40.0
#define SUMMARY_NUMBER_COLOR   64.0
#define SUMMARY_SCORE_COLOR   185.0
#define SUMMARY_BAR_COLOR
#define SUMMARY_BAR_ALPHA      0.65
#define SUMMARY_BAR_SPACING       8
#define SUMMARY_BAR_SIZE         10
#define SUMMARY_HEADER_1_SIZE
#define SUMMARY_HEADER_1_COLOR
#define SUMMARY_HEADER_2_SIZE
#define SUMMARY_HEADER_2_COLOR

#define HUD_LOCATION              0
#define HUD_HEIGHT              140
#define HUD_TEXT_START           10
#define HUD_TEXT_SCALE          3.0
#define HUD_GRADIENT_1       ORANGE
#define HUD_GRADIENT_2        BLANK

#define CONTROLS_COLOR_1      WHITE
#define CONTROLS_COLOR_2      BLACK
#define CONTROLS_TEXT_SCALE     3.0
#define CONTROLS_TEXT_START     9.0
#define CONTROLS_TEXT_SPACING   3.0

/// COLORS
///
#define PURE_GREEN       (Color){   0, 255,   0, 255 }
#define PURE_BLUE        (Color){   0,   0, 255, 255 }
#define PURE_YELLOW      (Color){ 255, 255,   0, 255 }

#define SCORE_PINK       (Color){ 242, 103, 119, 255 }
#define SCORE_ORANGE     (Color){ 255, 161,   0, 255 }
#define SCORE_GREEN      (Color){ 145, 240,   0, 255 }

#define WATER            (Color){  25, 113, 180, 255 }
#define SKY_BLUE         (Color){  25, 113, 180, 255 }
#define EYE_ORANGE       (Color){ 215, 115,   0, 255 }
#define CHILL_CYAN       (Color){  22, 230, 255, 255 }
#define KELP_GREEN       (Color){ 131, 231, 107, 255 }
#define KELP_BLUE        (Color){   0, 101, 203, 255 }
#define SKILL_PINK       (Color){ 235, 123, 217, 255 }
#define SHOCK_YELLOW     (Color){ 255, 197,  20, 255 }
#define SHOCK_RED        (Color){ 255,  41,  18, 255 }
#define ROOT_BORDER      (Color){  68,  56,  49, 255 }


/// CONVENIENCES
///
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))
#define TO_DEGREES(a) (a * (180.00 / PI))
#define TO_RADIANS(a) (a * (PI / 180.00))
#define UP                      0
#define DOWN                    1
#define LEFT                    2
#define RIGHT                   3


// This is a little janky but it reduces the need for a big external
// declaration file. Place the PULL_EXTERNS definition exactly
// once in a single source file (for simplicity I use the main file)
// and it'll include the header file and declare an instance
// of everything in here
//
// All other files will see this header as full of externs.
// Only downside is you won't be able to declare initial values
// in here.
//
#ifdef PULL_EXTERNS
    #define EXTERN
#else
    #define EXTERN extern
#endif



/// PROJECT_SPECIFIC
///
#define FIELD_X                214
#define FIELD_Y                100

#define MOVE_SCALE            4.00
#define TILE_SIZE            16.00
#define ACTOR_SCALE           4.00
#define GRAVITY_PULL          0.10
#define INFINITE             99999




/// Standard Assets
///
EXTERN Texture2D tex_32x;
//EXTERN Texture2D tex_64x;
EXTERN Texture2D box_tex;

EXTERN Texture2D tex_title;
EXTERN Texture2D tex_title_bg;
EXTERN Texture2D tex_summary_bg;

EXTERN Font hi_score_font;
EXTERN Font summary_font;
EXTERN Font controls_font;
EXTERN Font title_font;
EXTERN Font hud_font;

EXTERN Font font1;                               // Bubble font
EXTERN Font font3;                               // 3x5 font
EXTERN Font font4;                               // 7x6 Retro-future font
EXTERN Font font5;                               // Channel looking font

EXTERN RenderTexture2D screenbuf;                // For full-screen shader handling
EXTERN RenderTexture2D post_target;              // Secondary FBO used for post-processing

EXTERN fastPRNG::fastXS64 fastR;                 // fastPRNG, because Raylib's is not so good

EXTERN int sfx_buffer[16][2];                    // Allows buffering of sounds (up to 16, by clock ticks)
EXTERN int summary_data[16];                     // Holds copy of summary data for the end game screen

extern const char *main_options[];
extern const char *summary_headers[];
extern const char *summary_items[];
extern const char *summary_units[];

extern const char *menu_keyboard_tips[];
extern const char *menu_gamepad_tips[];
extern const char *game_keyboard_tips[];
extern const char *game_gamepad_tips[];



/// PROJECT-SPECIFIC ASSETS
///
EXTERN Texture2D tex_emoji_1;
EXTERN Texture2D tex_emoji_2;
EXTERN Texture2D tex_emoji_3;
EXTERN Texture2D tex_emoji_4;
EXTERN Texture2D tex_clouds;
EXTERN Texture2D tex_stars;
EXTERN Texture2D tex_aura;
EXTERN Texture2D tex_stripes;

EXTERN Sound sfx[MAX_SFX];
EXTERN Music muzak[MAX_MUSIC];

EXTERN char custom_text[80];
EXTERN bool hotkey_toggles[512];

extern Color colorset[16];



/// Declarations that require initialization
///
extern float anims[16];

extern int puppet;                               // Actor being controlled by the player
extern Camera2D camera;
extern Vector2 mouse;
extern Vector2 mouse_cam;

extern Shader scroll;
extern float scroll_vars[4];                     // Two sets of x/y offsets
extern int off_xLoc;
extern int off_yLoc;

extern Shader mosaic;
extern float m_var;
extern int mos_mLoc;
extern int mos_aLoc;

extern Shader wave;
extern float s_vars[6];                          // Brush sine wave
extern float l_vars[6];                          // Title emoji
extern float k_vars[6];                          // Kelp sine wave
extern float h_vars[6];                          // Hook sine wave
extern float f_vars[6];                          // Fire sine wave
extern int secondsLoc;
extern int freqXLoc;
extern int freqYLoc;
extern int ampXLoc;
extern int ampYLoc;
extern int speedXLoc;
extern int speedYLoc;

extern int loc_time;
extern int loc_strength;
extern float seconds;
extern float strength;

extern Shader crt;





/// STANDARD ROUTINES (MAIN)
///
extern void update_and_render();
extern void init_game_variables();


/// STANDARD ROUTINES (GAME_LOOP)
///
extern void core_game_loop();
extern void draw_control_hints(int mode);
extern void draw_title_menu(const char *msg);
extern void draw_title();
extern void draw_hud();
extern void draw_high_score();
extern void draw_summary(int mode);
extern void show_stats();
extern void end_game(int mode);


/// STANDARD ROUTINES (GAME_UPDATE)
///
extern void check_delta_updates();
extern void check_time();
extern void check_score();
extern void check_shaders();
extern void check_animations();
extern void check_actors();
extern void check_game_over();
extern void check_input();
extern void check_game_specific_anims();
extern void check_actors_while_paused();
extern void check_misc_logic();
extern void check_summary();


/// STANDARD ROUTINES (SPAWNS)
///
extern void spawn_permanent_actors();
extern int  spawn_actor(int type, float x, float y, float z, int cell_x, float path_x, float path_y, float path_z, bool force_reverse);
extern int  spawn_player(int x, int y, int z);






// Project-specific flags. Can wipe between projects (keep vars at top though)
//
struct flags_struct
{
 bool debug;                      // Enables debugging overlay
 bool tutorial;                   // TRUE if game is in tutorial mode
 bool basic_tutorial;             // TRUE if basic "panel-style" tutorial will be used
 bool tutorial_seen;              // TRUE if basic tutorial has been seen already
 bool game_over;                  // TRUE if "game over" conditions are met
 bool game_ending;                // TRUE if game ending conditions are met (animation before game over is set)
 bool stage_complete;             // TRUE if "stage complete" conditions are met
 bool paused;                     // TRUE if game logic is paused
 bool use_crt;                    // TRUE if CRT filter is enabled

 int total_actors;                // Total number of actors per-frame (debug var)
 int gamepad;                     // 1 if a gamepad is selected for primary use (else 0 for keyboard)
 int menu_pos;                    // Menu position (for cursor)
 int total_options;               // Menu position (for cursor)
 int game_mode;                   // 0 = title.    1 = Tutorial.     2 = game.    3 = Game Over.     4 = Credits?
 int game_stage;                  // Game stage (if game features multiple stages)
 int selected;                    // Value of player selection (if required)
 float aspect_ratio;              // Holds calculated screen aspect ratio
 float screenSize[2];             // Holds reported window size for passing to shaders

 float embers[64][3];             // holds data for burning embers over sacrifice gauge

 Color tint_col;                  // Full-screen tint control
 float tint_str;                  //
 float tint_velocity;             //

 float game_speed;                // Holds update rate for game (1.0 = normal, 2.0 = 2x speed, etc)
 float game_velocity;             //

 float mosaic;                    // Mosaic strength control (0-1f)
 float mosaic_velocity;           //

 int quake_duration;              // Quake frames left to animation

 int show_signal;                 // Triggers a banner across screen for x ticks (-1 to disable)
 int signal_time;                 // How many ticks signal has left on-screen (current)
 int signal_max_time;             // How many ticks signal has left on-screen (maximum)
 int signal_data[2];              // Shows optional data for the signal (i.e. 89723 BONUS!!)
 float signal_f_in;               // Fade in range for signal, based off total time
 float signal_f_out;              // Fade out range for signal, based off total time

 int high_score;                  // High score, if that feature is available
 int score;                       // Score
 int score_shake;                 // # of frames for score to shake (indicates a problem)
 int score_diff;                  // Amount to add into score over time

 float delta;                     // Delta-timing variables
 float delta_factor;              //
 float timing_leftover;           //
 int   last_updates;              //

 float game_time;                 // Elapsed game time
 int   clock_min;                 // Minutes on clock
 int   clock_sec;                 // Seconds on clock
 int   clock_sub;                 // Hundredths on clock

 #ifdef USE_LOG
     FILE *log_handle;
 #endif

 /// You can wipe everything below this point
 int covered;                     // Tracks how long 100% coverage has been seen
 int covered_countdown;           // Actual countdown
 int covered_sub;                 // Sub countdown

 int sum_launches;
 int sum_hits;
 int sum_escapes;
 int sum_yos;

 int rapid_fire;                  // Counts shots and reduces volume over time
 int shot_ready;                  // Tracks shot readiness (for player)
 int flip_dir;                    // Direction of flip (0 = none, 1 = flipped)
 int flip;                        // Timer for flip stuff

 int next_yo;                     // Score that will trigger the next YO! block
 float yo_factor;                 // Scoring factor required for a new YO block
};

EXTERN flags_struct flags;


#ifdef USE_LOG
    void open_log();
    void close_log();
    void log(const char *parser, ...);
#endif


#ifdef USE_TIPS
    #define MAX_TIPS 16

    struct tip_struct
    {
     bool active;
     int  msg_index;
     Rectangle window;
     Rectangle bounds;
    };

    EXTERN tip_struct tips[MAX_TIPS];

    extern void check_tips();
    extern int spawn_tip_holder(float x, float y, int _tip);
#endif




#ifdef USE_RINGS
    #define MAX_RINGS 32

    struct ring_struct
    {
     bool active;
     Vector2 pos;
     float radius;                            // Radius
     float r_change;                          // change in radius over time
     float alpha;                             // alpha (0-1)
     float a_change;                          // Change in alpha over time
     float thickness;                         // Thickness of line
     int segments;                            // # of segments
     int delay;                               // Delay before activation (allows concentric ring spawns at once)
     Color col;
    };

    EXTERN ring_struct rings[MAX_RINGS];

    extern void draw_rings();
    extern void check_rings();
    extern void init_ring(int slot);
    extern int get_free_ring();
    extern int spawn_ring(float x, float y, float radius, float thickness, int delay, Color col);
#endif




#ifdef USE_HIGH_SCORE
    struct hi_struct
    {
     int selected;                // Character currently selected (0-2 valid)
     unsigned char initials[3];   // Stores characters entered by user
     int blink;                   // True if we can display blinking chars
     int blink_timer;             // Timer to track above variable
    };

    EXTERN hi_struct hi_score;
    EXTERN unsigned char high_init[4];

    void write_high_score();
    int  read_high_score();
    void write_32bit_int(FILE *handle, int value);
    int  read_32bit_value(FILE *handle);
    int  get_initial();
#endif



#ifdef USE_BULLETIN
    struct bulletin_struct
    {
     bool enabled;                // Bulletin will be drawn if triggered
     bool position;               // 0 = top of screen   1 = bottom of screen

     int time_left;               // 'Ticks' before bulletin expires
     int height;                  // Height of bulletin in pixels
     int x;                       // X coordinate of scroll position
     int speed;                   // Scroll speed per tick
     int message;                 // Index of message (see bulletins array)

     Font font;                   // Font for bulletin

     float f_scale;               // Scaling value for font
     float f_spacing;             // Spacing value for font

     Color msg_col;               // Color of message
     Color banner_col;            // Color of background
    };

    extern bulletin_struct bulletin;

    extern const char *bulletins[];
    extern void draw_bulletin();
#endif



extern int spawn_actor(int type, float x, float y, float z, int cell_x, float path_x, float path_y, float path_z, bool force_reverse);
extern void kill_actor(int slot);




#ifdef USE_MATH
    float scale(float value, float source_min, float source_max, float scaled_min, float scaled_max);
    float tan_scale(float value, float source_min, float source_max, float scaled_min, float scaled_max);
    float log_scale(float value, float source_min, float source_max, float scaled_min, float scaled_max);

    float sin_march(float angle, float magnitude);
    float cos_march(float angle, float magnitude);

    float rescale(float input, float in_min, float in_max, float out_min, float out_max);
    float norm_f(float source, float min, float max);
    float FR_2d_angle_360_f(float x1, float y1, float x2, float y2);
    float degrees_to_radians(float degrees);

    int   FR_2d_dist(int x1, int y1, int x2, int y2);
    int   rotate_to_angle(int angle_a, int angle_b);

    Vector2 rotate_point_2d(float point_x, float point_y, float around_x, float around_y, float angle);
#endif





#ifdef USE_UTILITY
    bool valid_actor(int who);
    int get_free_actor(int mode);
    int get_population(int team);
    int get_radial_population(int who, int radius, int team);
    int is_odd(int number);

    void set_sound(int slot, float pitch, float volume);

    int quick_rand(int range);
    int check_floats(float one, float two, float range);
    int check_collision(int who, int target, int radius, int z_adjust);
    bool check_rect_collision(float check_x, float check_y, int x, int y, int w, int h, int diam_x, int diam_y);

    bool compare_color(Color one, Color two);
    Color makeacol(unsigned char r, unsigned char g, unsigned char b, unsigned char a);
    Color grab_gradient(Color c1, Color c2, float banding, float zero_to_one);
    void push_rect(Rectangle *rect, float x, float y, float w, float h);
    void justify_text(const char *text, float y, Font font, Color color, int scale, int spacing);
    void justify_text_xy(const char *text, float x, float y, Font font, Color color, int scale, int spacing);
    void justify_text_r(const char *text, float x, float y, Font font, Color color, int scale, int spacing);
    void justify_text_l(const char *text, float x, float y, Font font, Color color, int scale, int spacing);

    int dist_2d(int x1, int y1, int x2, int y2);
    int get_rough_dir(int x1, int y1, int x2, int y2);
    float get_angle_between(float x1, float y1, float x2, float y2);
    float actor_distance(int actor_1, int actor_2);

    Vector2 advance(float angle, float velocity);

    double easing_cubic(double x);
    float FR_easing_pop_in(float value, float source_min, float source_max, float scaled_min, float scaled_max, float pop_magnitude);

    // NB: These operate on IMAGES, not textures!
    //
    void blit(Image src, Image dest, int s_x, int s_y, int d_x, int d_y, int w, int h);
    void blit_h(Image src, Image dest, int s_x, int s_y, int d_x, int d_y, int w, int h);
    void blit_v(Image src, Image dest, int s_x, int s_y, int d_x, int d_y, int w, int h);

    // NB: These operate on TEXTURES
    //
    void tex_blit(Texture2D tex, int s_x, int s_y, int d_x, int d_y, int w, int h, float rotation, float scale, Color col);
    void tex_blit_f(Texture2D tex, int s_x, int s_y, int d_x, int d_y, int w, int h, float scale, float fade);
    void tex_blit_c(Texture2D tex, int s_x, int s_y, int d_x, int d_y, int w, int h, float scale, Color col);
    void tex_blit_h(Texture2D tex, int s_x, int s_y, int d_x, int d_y, int w, int h, float scale, Color col);
    void DrawTextureTiled(Texture2D texture, Rectangle source, Rectangle dest, Vector2 origin, float rotation, float scale, Color tint);

    void swap_int(int n1, int n2, int index);
    void quicksort(int low, int high);

    void open_comma_file(const char *filename);
    int parse_comma_file(unsigned char *buffer, bool first);

    void trigger_signal(int signal, int max_time, float fade_in_range, float fade_out_range);
#endif




#ifdef USE_ANIMATIONS
    struct anim_struct
    {
     int frame_x[MAX_FRAMES];     // x cell of frame inside sprite sheet (cell, not pixel)
     int frame_y[MAX_FRAMES];     // y cell of frame inside sprite sheet
     int timing[MAX_FRAMES];      // timing in game "ticks" to get to next frame
     int total_frames;            // maximum number of frames in this sequence
    };

    EXTERN anim_struct anim[MAX_ANIMATIONS];

    void generate_animations();
    void assign_animation(int who, int anim, int loops);
    void kill_animation(int who);
#endif




#ifdef USE_RTMIDI
    #include <iostream>
    #include <cstdlib>
    #include "RtMidi.h"

    #define MAX_KNOBS 17

    class device_struct {
        public:
        bool  mode;                                // 0 = Absolute, 1 = Relative
        int   id;                                  // MIDI code for this knob
        float r_min;                               // Raw minimum (input)
        float r_max;                               // Raw maximum (input)
        float s_min;                               // Scaled minimum (output)
        float s_max;                               // Scaled maximum (output)
        float value;                               // Scaled value for use in program
        int   input;                               // Raw value from MIDI controller
                                                   // This value persists if Relative
        float scale;                               // Used *ONLY* by Relative input to scale each tick
                        };

    class FR_midi_struct {
        public:
        bool ignore_sysex;                         // Set these flags to TRUE to ignore this message type
        bool ignore_timing;                        //
        bool ignore_active_sensing;                //

        bool keys[128];                            // Key state (pressed only)
        bool note_active;                          // **Primitive** variable for state of last key command
        int  note_key;                             // **primitive** variable for key # of last key command
        double note_freq;                          // **primitive** variable for frequency of last key command

        device_struct knobs[MAX_KNOBS];
        device_struct pads[MAX_KNOBS];
        int  total_ports;                          // Total MIDI ports, as reported by the lib
        int  current_port;                         // Port we're listening to. Defaults to zero.
        std::vector<unsigned char> message;        // Storage space for messages (later we'll queue this)
        int  msg_bytes;                            // Message byte length
        double msg_stamp;                          // Message timestamp delta
        RtMidiIn *midiin;                          // Obligatory rtMidi construct
                         };

    EXTERN FR_midi_struct midi;
    EXTERN Texture2D tex_knob;
    EXTERN Texture2D tex_knob_plate;

    void init_midi();
    void init_arturia_profile();
    void poll_midi();
    void draw_midi_knobs();
    void update_midi_knob(int slot);
    void set_knob(int slot, float new_value);
    void save_knobs();
    void load_knobs();
    void rescale_midi_device_abs(device_struct &knob, float raw_min, float raw_max, float scale_min, float scale_max);
    void rescale_midi_device_rel(device_struct &knob, float scale_min, float scale_max, float tick_scale);
#endif





#ifdef USE_TEXTBOX
    #define MAX_TEXTBOX          16

    struct textbox_struct
    {
     Rectangle pos;
     float scale;                 // Internally-used window scale variable
     float text_scale;            // Raylib text scale factor
     bool enabled;                // TRUE if this box is active
     bool hold;                   // Holds textbox in open state (can be tied to external logic)
     int viewable_text;           // Number of chars currently exposed (starts at zero, counts up)
     int duration;                // Number of ticks before textbox closes. -1 if disabled
     int mode;                    // 0 = entering, 1 = displaying text, 2 = exiting
     int ticks;                   // Counts time between displaying characters
     const char *msg;
    };

    extern textbox_struct textbox[MAX_TEXTBOX];

    extern float box_tex_x;                // For anchoring happy texture at game over summary
    extern float box_tex_y;                //
    extern int box_update;                 //
    extern const char *textbox_messages[];

    void check_textbox();
    void draw_text_box(int slot);
    void trigger_textbox(int slot, int x, int y, int w, int h, int message, int delay, int duration);
#endif






struct actor_struct
{
 Texture2D cel;         // Pointer to source texture
 Vector3 pos;           // Position
 Vector3 offset;        // Destination (only used to move enemies around)
 Vector3 dest;          // Destination (only used to move enemies around)
 Vector3 path;          // (OPTIONAL) Path of travel (x/y/z)
 Vector3 check;         // (OPTIONAL) Internal variable for checking collisions
 Color color;           // Color tint (WHITE by default)
 Color ref_color;       // Reference Color (if you need to backup original color)

 float blend;           // Blend strength (0-1)    < only applies if not solid >
 float angle;           // Angle of rotation (0-360)
 float v_angle;         // (OPTIONAL) Angle of movement (0-360)
 float v_speed;         // (OPTIONAL) Velocity of movement (tied to v_angle)
 float v_roll;          // (OPTIONAL) != 0 if actor rolls while in flight
 float slew;            // (OPTIONAL) != 0 if actor slides on X axis in flight

 float scale;           // Actor scale (populated during projection)
 float roll;            // != 0 if actor's sprite rolls in place
 float y_shift;         // Dirty hack to physically shift a sprite on y axis by ? pixels (pre-projection)

 float quake;           // > 0 if actor is rolling from damage
 float speed;           // Actor's base speed (0.0 to 0.20) which is added to their movement each turn
 float last_angle;      // Last angle actor had before rotation command (for squid)

 int team;              // 1 = Friendly    2 = Enemy    6 = Inert substances
 int shadow;            // > -1 if an actor is their shadow
 int _size_w;           // Dimensions of the cel frame
 int _size_h;           // Dimensions of the cel frame
 int flip;              // Force horizontal mirroring
 int anim_set;          // Animation set to reference
 int anim_step;         // Which step in the anim sequence we're at
 int anim_timing;       // Current step timing for animation
 int anim_loop;         // # of times to loop animation. 0 if only played once
 int mode;              // Used for misc tracking (per-project)
 int home[3];           // Home information (if an origin needs defining)
 int state;             // 0 = inactive. 1-12 = doing stuff
 int effect;            // 0 = none.
 int jitter_x;          // > 0 if actor randomly jitters by 1 PIX +/- on x axis
 int jitter_y;          // > 0 if actor randomly jitters by 1 PIX +/- on y axis
 int owner;             // Actor that owns this (used to define simple relationships)
 int timer;             // Internal timer for animation purposes, or whatever you need
 int stamina;           // Walking and being attacked drain stamina. When <= 0 the actor dies or fades out
 int cel_x;             // X cel within sprite sheet
 int cel_y;             // Y cel within sprite sheet
 int preference;        // Extra data for specifying preferred behavior(s)
 int pref_delay;        // Time before actor can be influenced again
 int locked_to;         // Actor this player was triggered by (so ammo doesn't hit the actor firing it!)
 int following;         // Actor this actor is following
 int follow_x;          // X Pixel offset for follower
 int follow_y;          // Y Pixel offset for follower
 int extra[4];          // Extra data points for use wherever (lazy!)
 int event;             // > 0 if actor will trigger an event when their commitment has ended
 int inert;             // > 0 if actor should not check collisions. Count down.
 int tip;               // > -1 if actor has a tip attached to them

 int updates;           // Number of movement updates per -tick (default is 1)
 int moved_up;          // > 0 if actor has moved up recently
 int moved_down;        // > 0 if actor has moved down recently
 int moved_wait;        // > 0 if actor needs to wait before heading to next way-point

 bool d_repeat[4];      // Tag for direction repeat preclusion
 bool b_repeat[4];      // Tag for button repeat preclusion
 int d_pad[4];          // Janky virtual gamepad (refine later)
 int buttons[4];        //

 int shader;            // > 0 if a shader is in use with this actor
 float s_vars[10];      // Storage space for shader vars. These will vary according to the shader in use

 bool immortal;         // TRUE is actor is immune to death
 bool decal;            // Indicates actor is a floor decal, and will be drawn before normal actors
 bool flicker;          // Tracks actor flicker (used by effects)
 bool mirror;           // TRUE if actor is mirrored on X-axis
 bool gravity;          // TRUE if gravity affects Y coordinate
 bool launched;         // TRUE if player has been launched (do not repeat action until land)
 bool projected;        // TRUE is actor is passed through projection stage before drawing
 bool enabled;          // TRUE if this actor is active
 bool visible;          // TRUE if this actor will be drawn to screen
 bool accessory;        // TRUE if actor will be rendered with a package object
                        // and should not be drawn or interacted with otherwise
};

EXTERN actor_struct actors[MAX_ACTORS];
EXTERN int sorted[MAX_ACTORS][2];




#ifdef USE_FM_SYNTH
    #define fm_PI 3.14159265f
    #define FM_MONO_INPUT                         // Comment this out if STEREO input is necessary

    // Contains FM wave form settings
    //
    class FM_form {
      public:

      bool  filter_on; // unused, but stored in .sfs files
      int   wave_type;
      float base_freq;
      float freq_limit;
      float freq_ramp;
      float freq_dramp;
      float duty;
      float duty_ramp;
      float vib_strength;
      float vib_speed;
      float vib_delay;
      float env_attack;
      float env_sustain;
      float env_decay;
      float env_punch;
      float lpf_resonance;
      float lpf_freq;
      float lpf_ramp;
      float hpf_freq;
      float hpf_ramp;
      float pha_offset;
      float pha_ramp;
      float repeat_speed;
      float arp_speed;
      float arp_mod;
                  };




    // Contains misc FM synth settings
    //
    class FM_synth {
      public:

      float  master_vol;
      float  sound_vol;
      bool   playing_sample;
      int    phase;
      double fperiod;
      double fmaxperiod;
      double fslide;
      double fdslide;
      int    period;
      float  square_duty;
      float  square_slide;
      int    env_stage;
      int    env_time;
      int    env_length[3];
      float  env_vol;
      float  fphase;
      float  fdphase;
      int    iphase;
      float  phaser_buffer[1024];
      int    ipp;
      float  noise_buffer[32];
      float  fltp;
      float  fltdp;
      float  fltw;
      float  fltw_d;
      float  fltdmp;
      float  fltphp;
      float  flthp;
      float  flthp_d;
      float  vib_phase;
      float  vib_speed;
      float  vib_amp;
      int    rep_time;
      int    rep_limit;
      int    arp_time;
      int    arp_limit;
      double arp_mod;
      int    wav_bits;
      int    wav_freq;
      int    file_sampleswritten;
      float  filesample;
      int    fileacc;
                   };

    // From FR_FM_SYNTH -
    //
    extern bool FM_save_form(const char *filename);
    extern bool FM_load_form(const char *filename);
    extern void FM_reset_sample(bool restart);
    extern void FM_load_assets();
    extern void FM_reset_form();
    extern void FM_play_sample();
    extern Sound FM_export_to_Sound();
#endif




/// Project-specific ----------------------------------------------------------- ///

EXTERN int pixel_field[FIELD_Y][FIELD_X];
EXTERN int barrier_field[FIELD_Y][FIELD_X];
EXTERN Texture2D tex_container;
EXTERN Texture2D tex_moai;


/// PROJECT-SPECIFIC ROUTINES
///
int  spawn_radiance(float x, float y, float z, bool dir);
int  spawn_enemy(float x, float y, float z, int color, int direction, float wave_offset, int variety);
int  spawn_brown_bomber_block();
int  spawn_pixel(int source_actor);
int  spawn_dirt(int source_actor);
int  spawn_antimatter(int source_actor);
int  spawn_yo_block();
int  spawn_moai(int container);
int  spawn_rocket_flame(int moai);
int  spawn_debris(int attacker, int victim);
int  spawn_remnant(int f_x, int f_y, int color_set);
void spawn_wave();
void destroy_enemy(int target, int attacker);
void draw_pixel_field();
void draw_warning();
void draw_containers();
void draw_sky();
bool field_collision(int who);
bool antimatter_collision(int who);
void update_field();
void clear_field(int x, int y, int w, int h, bool protect_filling);
void clear_field_color(int color);
void mark_field_containers();
bool is_clear(int x, int y, int permeable);
void TEST_fill_container(int which);
bool is_container_full(int container);
bool is_game_over();
void launch_moai(int container);
void fire_shot(float x, float y, float z, int direction);
void kill_player();
